"""This module contains internal classes that store matched variants and format them
to be compatible with plink2 --score
"""

import collections.abc
import gzip
import pathlib

from ._match.plink import plinkify, pivot_score


class PlinkFrame:
    """A PlinkFrame contains matches and makes them suitable for plink2 --score

    A PlinkFrame represents the largest possible set of variants that can be written
    to one plink score file. Including other effect types breaks dosage calculation.
    Including duplicate IDs with different effect alleles (n) breaks plink2 --score
    When writing to a file the set of variants can be split into separate files (one
    per chromosome) for larger datasets.
    """

    def __init__(self, effect_type, n, df):
        self.effect_type = effect_type
        self.n = n
        self.df = df

    def __repr__(self):
        return f"{type(self).__name__}(effect_type={self.effect_type!r}, n={self.n!r}, df={self.df!r})"

    def split_pivot(self):
        """Splitting scoring files is helpful for split - apply - combine on big data"""
        dfs = self.df.collect().partition_by(["chr_name"], as_dict=True)
        return {k[0]: v.pipe(pivot_score) for k, v in dfs.items()}

    def pivot_wide(self):
        """Pivoting wide is important to enable parallel score calculation"""
        return self.df.collect().pipe(pivot_score)

    def write(self, directory, dataset, split=False):
        """Write a plink2 --score compatible file, optionally splitting"""
        fouts = []
        if split:
            dfs = self.split_pivot()
            for chrom, df in dfs.items():
                fout = (
                    pathlib.Path(directory)
                    / f"{dataset}_{chrom}_{str(self.effect_type)}_{self.n}.scorefile.gz"
                )
                with gzip.open(fout, "wb") as f:
                    df.write_csv(f, separator="\t")
                fouts.append(fout)
        else:
            chrom = "ALL"
            fout = (
                pathlib.Path(directory)
                / f"{dataset}_{chrom}_{str(self.effect_type)}_{self.n}.scorefile.gz"
            )
            df = self.pivot_wide()
            with gzip.open(fout, "wb") as f:
                df.write_csv(f, separator="\t")
            fouts.append(fout)
        return fouts


class PlinkFrames(collections.abc.Sequence):
    """A container of PlinkFrames"""

    def __init__(self, *elements):
        self._elements = list(elements)

    def __getitem__(self, item):
        return self._elements[item]

    def __len__(self):
        return len(self._elements)

    def __repr__(self):
        return f"{type(self).__name__}({self._elements!r})"

    @classmethod
    def from_matchresult(cls, matchresult):
        effect_types = plinkify(matchresult)
        plinkframes = []
        for effect_type, dataframes in effect_types.items():
            for i, df in enumerate(dataframes):
                df = df.collect()
                if not df.is_empty():
                    plinkframes.append(
                        PlinkFrame(effect_type=effect_type, n=i, df=df.lazy())
                    )

        return cls(*plinkframes)



"""This module labels match candidates with various flag columns, with the aim of
producing a final set of best match candidates (one maximum for each variant).

These operations are all quite data-framey and involve calculating a horizontal
maximum across boolean columns to determine if a variant should be given a true value
in an exclude column.
"""

import logging
import pathlib

import polars as pl
from xopen import xopen

from .preprocess import complement_valid_alleles

logger = logging.getLogger(__name__)


def label_matches(
    df: pl.LazyFrame,
    keep_first_match,
    remove_ambiguous,
    remove_multiallelic,
    skip_flip,
    filter_IDs,
) -> pl.LazyFrame:
    """Label match candidates with additional metadata. Column definitions:

    - match_candidate: All input variants that were returned from match.get_all_matches() (always True in this function)
    - best_match: True if row is the best possible match type (refalt > altref > ...)
    - duplicate: True if more than one best match exists for the same accession and ID
    - ambiguous: True if ambiguous
    """
    labelled = (
        df.with_columns(
            exclude=pl.lit(False)
        )  # set up dummy exclude column for _label_*
        .pipe(_label_best_match)
        .pipe(_label_duplicate_best_match)
        .pipe(_label_duplicate_id, keep_first_match)
        .pipe(_label_biallelic_ambiguous, remove_ambiguous)
        .pipe(_label_multiallelic, remove_multiallelic)
        .pipe(_label_flips, skip_flip)
        .pipe(_label_filter, filter_IDs)
        .with_columns(match_candidate=pl.lit(True))
    )

    return labelled.pipe(_encode_match_priority)


def _encode_match_priority(df: pl.LazyFrame) -> pl.LazyFrame:
    """Encode a new column called match status containing matched, unmatched, excluded, and not_best"""
    return (
        df.with_columns(
            # set false best match to not_best
            match_priority=pl.col("best_match").map_elements(
                lambda x: {None: 0, True: 1, False: 3}[x], return_dtype=pl.UInt8
            )
        )
        .with_columns(
            excluded_match_priority=pl.col("exclude").map_elements(
                lambda x: {None: 0, True: 2, False: 0}[x], return_dtype=pl.UInt8
            )
        )
        .with_columns(
            max=pl.max_horizontal("match_priority", "excluded_match_priority")
        )
        .with_columns(
            match_status=pl.col("max")
            .map_elements(
                lambda x: {0: "unmatched", 1: "matched", 2: "excluded", 3: "not_best"}[
                    x
                ],
                return_dtype=pl.String,
            )
            .cast(pl.Categorical)
        )
        .drop(["max", "excluded_match_priority", "match_priority"])
    )


def _label_best_match(df: pl.LazyFrame) -> pl.LazyFrame:
    """Best matches have the lowest match priority type. Find the best matches and label them."""
    logger.debug("Labelling best match type (refalt > altref > ...)")
    match_priority = {
        "refalt": 0,
        "altref": 1,
        "refalt_flip": 2,
        "altref_flip": 3,
        "no_oa_ref": 4,
        "no_oa_alt": 5,
        "no_oa_ref_flip": 6,
        "no_oa_alt_flip": 7,
    }
    match_priority_rev = {v: k for k, v in match_priority.items()}

    # use a groupby aggregation to guarantee the number of rows stays the same
    # rows were being lost using an anti join + reduce approach
    prioritised: pl.LazyFrame = (
        df.with_columns(match_priority=pl.col("match_type").replace(match_priority))
        .with_columns(
            best_match_type=pl.col("match_priority")
            .min()
            .over(["accession", "row_nr"])
            .replace(match_priority_rev)
        )
        .with_columns(
            best_match=pl.when(pl.col("best_match_type") == pl.col("match_type"))
            .then(pl.lit(True))
            .otherwise(pl.lit(False))
        )
    )

    return prioritised.drop(["match_priority", "best_match_type"])


def _label_duplicate_best_match(df: pl.LazyFrame) -> pl.LazyFrame:
    """A scoring file row_nr in an accession group can be duplicated if a target position has different REF, e.g.:

    ┌────────┬────────────────────────┬────────────┬────────────────┬─────┬────────────┐
    │ row_nr ┆ accession              ┆ match_type ┆ ID             ┆ REF ┆ best_match │
    │ ---    ┆ ---                    ┆ ---        ┆ ---            ┆ --- ┆ ---        │
    │ i64    ┆ cat                    ┆ str        ┆ cat            ┆ str ┆ bool       │
    ╞════════╪════════════════════════╪════════════╪════════════════╪═════╪════════════╡
    │ 38557  ┆ PGS000012_hmPOS_GRCh37 ┆ no_oa_alt  ┆ 3:29588979:A:G ┆ A   ┆ true       │
    ├╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┤
    │ 38557  ┆ PGS000012_hmPOS_GRCh37 ┆ no_oa_alt  ┆ 3:29588979:T:G ┆ T   ┆ true       │
    └────────┴────────────────────────┴────────────┴────────────────┴─────┴────────────┘

    Label the first row with best_match = true, and duplicate rows with best_match = false
    """
    logger.debug(
        "Labelling duplicated best match: keeping first instance as best_match = True"
    )
    labelled: pl.LazyFrame = (
        df.with_columns(
            count=pl.col("best_match").len().over(["accession", "row_nr", "best_match"])
        )
        .with_columns(
            duplicate_best_match=pl.when((pl.col("count") > 1) & (pl.col("best_match")))
            .then(pl.lit(True))
            .otherwise(pl.lit(False))
        )
        .drop("count")
        .with_row_index(
            name="temp_row_nr"
        )  # add temporary row count to get first variant
        .with_columns(
            best_match=pl.when(
                pl.col("best_match")
                & pl.col("duplicate_best_match")
                & (pl.col("temp_row_nr") > pl.min("temp_row_nr")).over(
                    ["accession", "row_nr"]
                )
            )
            .then(False)  # reset best match flag for duplicates
            .otherwise(pl.col("best_match"))  # just keep value from existing column
        )
    )

    return labelled


def _label_duplicate_id(df: pl.LazyFrame, keep_first_match: bool) -> pl.LazyFrame:
    """Label best match duplicates made when the scoring file is remapped to a different genome build

    ┌─────────┬────────────────────────┬─────────────┬────────────────┬─────┬────────────┐
    │ row_nr  ┆ accession              ┆ match_type  ┆ ID             ┆ REF ┆ best_match │
    │ ---     ┆ ---                    ┆ ---         ┆ ---            ┆ --- ┆ ---        │
    │ i64     ┆ cat                    ┆ str         ┆ cat            ┆ str ┆ bool       │
    ╞═════════╪════════════════════════╪═════════════╪════════════════╪═════╪════════════╡
    │ 1194115 ┆ PGS002244_hmPOS_GRCh37 ┆ altref      ┆ 3:50924580:C:A ┆ C   ┆ true       │
    ├╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┤
    │ 1194132 ┆ PGS002244_hmPOS_GRCh37 ┆ refalt_flip ┆ 3:50924580:C:A ┆ C   ┆ true       │
    └─────────┴────────────────────────┴─────────────┴────────────────┴─────┴────────────┘

    refalt > altref > ... prioritisation doesn't fix this problem because row_nr is different (duplicated by remapping)
    """
    logger.debug(
        "Labelling multiple scoring file lines (accession/row_nr) that best_match to the same variant"
    )

    # the window in .over() starts with accession + ID
    # best_match is added to not count: same row_nr, different match_type (_label_best_match)
    # duplicate_best_match is added to not count: same row_nr, same match_type (_label_duplicate_best_match)
    duplicates = df.with_columns(
        count=pl.col("ID")
        .len()
        .over(["accession", "ID", "best_match", "duplicate_best_match"])
    ).with_columns(
        duplicate_ID=pl.when((pl.col("count") > 1) & (pl.col("best_match")))
        .then(pl.lit(True))
        .otherwise(pl.lit(False))
    )

    if keep_first_match:
        logger.debug("Keeping first duplicate, labelling others with exclude flag ")
        # set first duplicate (with the smallest row_nr) to exclude = false
        labelled = duplicates.with_columns(
            exclude_duplicate=pl.when(
                (pl.col("duplicate_ID"))
                & (
                    pl.col("row_nr")
                    != pl.min("row_nr").over(["accession", "ID", "duplicate_ID"])
                )
            )
            .then(True)
            .otherwise(False)
        )
    else:
        logger.debug("Labelling all duplicates with exclude flag")
        labelled = duplicates.with_columns(exclude_duplicate=pl.col("duplicate_ID"))

    # get the horizontal maximum to combine the exclusion columns for each variant
    return (
        labelled.with_columns(max=pl.max_horizontal("exclude", "exclude_duplicate"))
        .drop(["exclude", "exclude_duplicate"])
        .rename({"max": "exclude"})
    )


def _label_biallelic_ambiguous(df: pl.LazyFrame, remove_ambiguous) -> pl.LazyFrame:
    logger.debug("Labelling ambiguous variants")
    ambig = (
        df.with_columns(
            [
                pl.col(
                    [
                        "effect_allele",
                        "other_allele",
                        "REF",
                        "ALT",
                        "effect_allele_FLIP",
                        "other_allele_FLIP",
                    ]
                ).cast(str),
                pl.lit(True).alias("ambiguous"),
            ]
        ).pipe(complement_valid_alleles, ["REF"])
    ).with_columns(
        pl.when(pl.col("REF_FLIP") == pl.col("ALT"))
        .then(pl.col("ambiguous"))
        .otherwise(False)
    )

    if remove_ambiguous:
        logger.debug("Labelling ambiguous variants with exclude flag")
        return (
            ambig.with_columns(
                pl.when(pl.col("ambiguous"))
                .then(True)
                .otherwise(False)
                .alias("exclude_ambiguous")
            )
            .with_columns(max=pl.max_horizontal("exclude", "exclude_ambiguous"))
            .drop(["exclude", "exclude_ambiguous"])
            .rename({"max": "exclude"})
        )
    else:
        return (
            ambig.with_column(pl.lit(False).alias("exclude_ambiguous"))
            .with_columns(max=pl.max_horizontal("exclude", "exclude_ambiguous"))
            .drop(["exclude", "exclude_ambiguous"])
            .rename({"max": "exclude"})
        )


def _label_multiallelic(df: pl.LazyFrame, remove_multiallelic: bool) -> pl.LazyFrame:
    """Label multiallelic variants with exclude flag

    (Multiallelic variants are already labelled with the "is_multiallelic" column in match.preprocess)
    """
    if remove_multiallelic:
        logger.debug("Labelling multiallelic matches with exclude flag")
        return df.with_columns(
            pl.when(pl.col("is_multiallelic"))
            .then(True)
            .otherwise(pl.col("exclude"))  # don't overwrite existing exclude flags
            .alias("exclude")
        )
    else:
        logger.debug("Not excluding multiallelic variants")
        return df


def _label_flips(df: pl.LazyFrame, skip_flip: bool) -> pl.LazyFrame:
    df = df.with_columns(
        pl.when(pl.col("match_type").str.contains("_flip"))
        .then(True)
        .otherwise(False)
        .alias("match_flipped")
    )
    if skip_flip:
        logger.debug("Labelling flipped matches with exclude flag")
        return df.with_columns(
            pl.when(pl.col("match_flipped"))
            .then(True)
            .otherwise(pl.col("exclude"))  # don't overwrite existing exclude flags
            .alias("exclude")
        )
    else:
        logger.debug("Not excluding flipped matches")
        return df.with_columns(match_IDs=pl.lit("NA"))


def _label_filter(df: pl.LazyFrame, filter_IDs: pathlib.Path) -> pl.LazyFrame:
    if filter_IDs is not None:
        logger.debug("Reading filter file (variant IDs)")
        with xopen(filter_IDs, "r") as f:
            filt_series = pl.Series([line.strip() for line in f], dtype=pl.Utf8)

        nIDs = len(filt_series)
        logger.debug(
            "Excluding variants that are not in ID list (read {} IDs)".format(nIDs)
        )
        df = df.with_columns(pl.col("ID").is_in(filt_series).alias("match_IDs"))
        return df.with_columns(
            pl.when(~pl.col("match_IDs"))
            .then(True)
            .otherwise(pl.col("exclude"))
            .alias("exclude")
        )
    else:
        logger.info("--filter_IDs not set, skipping filtering")
        return df



import logging

import polars as pl

from pgscatalog.core import EffectType

logger = logging.getLogger(__name__)


def plinkify(df):
    """Prepare dataframe to align with plink2 expectations for a --scorefile

    Split effect types: weight calculation differs by effect type
    Variants with the same ID and different effect alleles must be split into different files

    Input df must contain match candidates filtered to best matched variant
    """
    min_cols = [
        "accession",
        "effect_type",
        "chr_name",
        "ID",
        "matched_effect_allele",
        "effect_weight",
    ]
    df = df.select(min_cols).lazy()

    # 1. split by effect type
    effect_frames = zip(
        [EffectType.ADDITIVE, EffectType.DOMINANT, EffectType.RECESSIVE],
        _split_effect_type(df),
        strict=True,
    )

    # 2. deduplicate
    deduped = {}
    for effect_type, effect_frame in effect_frames:
        deduped.setdefault(effect_type, [])
        deduped[effect_type].append(effect_frame)

    return deduped


def pivot_score(df: pl.DataFrame) -> pl.DataFrame:
    """Format a dataframe to plink2 --score standard
    Minimum example:
    ID | effect_allele | effect_weight
    Multiple scores are OK too:
    ID | effect_allele | weight_1 | ... | weight_n
    """
    # fill_null: value = 0 not strategy = "zero"
    # we always handle effect weights as strings because we never modify them
    return (
        df.pivot(
            index=["ID", "matched_effect_allele", "effect_type"],
            values="effect_weight",
            columns="accession",
        )
        .rename({"matched_effect_allele": "effect_allele"})
        .fill_null(value="0")
        .drop("effect_type")
    )


def _check_column_types(matches: pl.LazyFrame):
    logger.debug("Checking column types")
    # these columns are most important for writing out
    correct_schema = {
        "chr_name": pl.Categorical,
        "chr_position": pl.UInt64,
        "ID": pl.Utf8,
        "matched_effect_allele": pl.Categorical,
        "effect_weight": pl.Float64,
        "effect_type": pl.Categorical,
        "accession": pl.Categorical,
    }
    col_types = {
        x: matches.schema.get(x)
        for x in list((matches.schema.keys() & correct_schema.keys()))
    }
    if not (col_types == correct_schema):
        logger.critical(
            f"MISMATCHED SCHEMA\nCurrent columns: {col_types}\nCorrect schema:{correct_schema}"
        )
        raise ValueError


def _split_effect_type(
    df: pl.LazyFrame,
) -> tuple[pl.LazyFrame, pl.LazyFrame, pl.LazyFrame]:
    additive = df.filter(pl.col("effect_type") == "additive")
    dominant = df.filter(pl.col("effect_type") == "dominant")
    recessive = df.filter(pl.col("effect_type") == "recessive")
    return additive, dominant, recessive


def _deduplicate_variants(df: pl.LazyFrame) -> list[pl.LazyFrame | None]:
    """Find variant matches that have duplicate identifiers
    When merging a lot of scoring files, sometimes a variant might be duplicated
    this can happen when the matched effect allele differs at the same position, e.g.:
        - chr1: chr2:20003:A:C A 0.3 NA
        - chr1: chr2:20003:A:C C NA 0.7
    where the last two columns represent different scores.  plink demands
    unique identifiers! so need to split, score, and sum later
    Parameters:
    df: A dataframe containing all matches, with columns ID, effect_allele, and
        effect_weight
    Returns:
        A list of dataframes, with unique ID - matched effect allele combinations
    """
    if df.select("ID").head().collect().is_empty():
        logger.info("Empty input: skipping deduplication")
        return [None]
    else:
        logger.debug("Deduplicating variants")

    # 1. unique ID - EA is important because normal duplicates are already
    #   handled by pivoting, and it's pointless to split them unnecessarily
    # 2. use cumcount to number duplicate IDs
    # 3. join cumcount data on original DF, use this data for splitting
    # note: effect_allele should be equivalent to matched_effect_allele
    ea_count: pl.LazyFrame = (
        df.select(["ID", "matched_effect_allele"])
        .unique()
        .with_columns(
            [
                pl.col("ID").cumcount().over(["ID"]).alias("cumcount"),
                pl.col("ID").count().over(["ID"]).alias("count"),
            ]
        )
    )

    dup_label: pl.LazyFrame = df.join(
        ea_count, on=["ID", "matched_effect_allele"], how="left"
    )

    # now split the matched variants, and make sure we don't lose any
    # cumcount = ngroup-1, so add 1
    n_splits: int = ea_count.select("cumcount").max().collect().item(0, 0) + 1
    df_lst: list = []

    for i in range(0, n_splits):
        x: pl.LazyFrame = dup_label.filter(pl.col("cumcount") == i).drop(
            ["cumcount", "count"]
        )
        if (df := x.collect()).is_empty():
            continue
        else:
            df_lst.append(df.lazy())

    return df_lst



import logging
import os
import shutil

import polars as pl

from pgscatalog.core import NormalisedScoringFile

from ._arrow import loose
from ._match.preprocess import complement_valid_alleles
from ._match.match import get_all_matches
from .matchresult import MatchResult

logger = logging.getLogger(__name__)


def match_variants(score_df, target_df, target):
    """Get all match candidates for a VariantFrame dataframe and ScoringFileFrame dataframe

    Returns a :class:`MatchResult`
    """
    matches = get_all_matches(scorefile=score_df, target=target_df)
    return MatchResult(dataset=target.dataset, matchresult=matches)


class ScoringFileFrame:
    """Like :class:`pgscatalog.core.NormalisedScoringFile`, but backed by the polars dataframe library

    Instantiated with a :class:`pgscatalog.core.NormalisedScoringFile` written to
    a file. This is a long format/melted CSV file containing normalised variant data
    (i.e. the output of combine scorefiles application):

    >>> from ._config import Config
    >>> path = Config.ROOT_DIR.parent / "pgscatalog.core" / "tests" / "data" / "combined.txt.gz"
    >>> x = ScoringFileFrame(path)
    >>> x  # doctest: +ELLIPSIS
    ScoringFileFrame(NormalisedScoringFile('.../combined.txt.gz'))

    Using a context manager is important to prepare a polars dataframe:

    >>> with x as arrow:
    ...     assert all(os.path.exists(x) for x in x.arrowpaths)
    ...     arrow.collect().shape
    (154, 11)
    >>> assert not any(os.path.exists(x) for x in x.arrowpaths)  # all cleaned up

    >>> from .variantframe import VariantFrame
    >>> path = Config.ROOT_DIR.parent / "pgscatalog.core" / "tests" / "data" / "hapnest.bim"
    >>> target = VariantFrame(path, dataset="hapnest")
    >>> with target as target_df, x as score_df:
    ...     match_variants(score_df=score_df, target_df=target_df, target=target)  # doctest: +ELLIPSIS
    MatchResult(dataset=hapnest, matchresult=[<LazyFrame ...
    """

    def __init__(self, path, chrom=None, cleanup=True, tmpdir=None):
        self.scoringfile = NormalisedScoringFile(path)
        # used for filtering the scoring file if the target contains a single chromosome
        self.chrom = chrom
        self._cleanup = cleanup
        self._tmpdir = tmpdir
        self._loosed = False
        self.arrowpaths = None

    def __enter__(self):
        """Use a context manager to create arrow IPC files and return a lazyframe"""
        # prevent nested context managers creating multiple arrow files
        pl.enable_string_cache()

        if not self._loosed:
            logger.debug(f"Converting {self!r} to feather format")
            self.arrowpaths = loose(
                self.scoringfile,
                tmpdir=self._tmpdir,
            )
            self._loosed = True
            logger.debug(f"{self!r} feather conversion complete")

        if self.arrowpaths is not None:
            score_df = (
                pl.scan_ipc(self.arrowpaths)
                .pipe(
                    complement_valid_alleles,
                    flip_cols=["effect_allele", "other_allele"],
                )
                .with_columns(
                    [
                        pl.col("chr_name").cast(pl.Categorical),
                        pl.col("accession").cast(pl.Categorical),
                        pl.col("effect_allele").cast(pl.Categorical),
                        pl.col("other_allele").cast(pl.Categorical),
                        pl.col("effect_allele_FLIP").cast(pl.Categorical),
                        pl.col("other_allele_FLIP").cast(pl.Categorical),
                    ]
                )
            )
        else:
            raise ValueError("No arrow IPC file")

        if self.chrom is not None:
            # add filter to query plan
            logger.debug(f"Filtering scoring file to chromosome {self.chrom}")
            score_df = score_df.filter(pl.col("chr_name") == self.chrom)

        return score_df

    def __exit__(self, *args, **kwargs):
        """Optionally clean up the arrow IPC files"""
        if self._cleanup:
            [os.unlink(x) for x in self.arrowpaths]
            self._loosed = False
        pl.disable_string_cache()

    def __repr__(self):
        return f"{type(self).__name__}({repr(self.scoringfile)})"

    def save_ipc(self, destination):
        """Save the dataframe prepared by the context manager to an Arrow IPC file

        Useful because the context manager will clean up the IPC files while exiting.

        This method allows data to be persisted."""
        if not self._loosed:
            raise ValueError(
                "Can't save IPC because it doesn't exist."
                "Try calling inside a with: statement"
            )

        shutil.copyfile(self.arrowpaths, destination)



import logging

import polars as pl

logger = logging.getLogger(__name__)


def get_all_matches(
    scorefile: pl.LazyFrame, target: pl.LazyFrame
) -> list[pl.LazyFrame]:
    scorefile_oa = scorefile.filter(pl.col("other_allele").is_not_null())
    scorefile_no_oa = scorefile.filter(pl.col("other_allele").is_null())

    matches: list[pl.LazyFrame()] = []
    col_order = [
        "row_nr",
        "chr_name",
        "chr_position",
        "effect_allele",
        "other_allele",
        "effect_weight",
        "effect_type",
        "accession",
        "effect_allele_FLIP",
        "other_allele_FLIP",
        "ID",
        "REF",
        "ALT",
        "is_multiallelic",
        "matched_effect_allele",
        "match_type",
    ]

    logger.debug("Getting matches for scores with effect allele and other allele")
    matches.append(
        match_variants(
            scorefile=scorefile_oa, target=target, match_type="refalt"
        ).select(col_order)
    )
    matches.append(
        match_variants(scorefile_oa, target, match_type="altref").select(col_order)
    )
    matches.append(
        match_variants(scorefile_oa, target, match_type="refalt_flip").select(col_order)
    )
    matches.append(
        match_variants(scorefile_oa, target, match_type="altref_flip").select(col_order)
    )

    logger.debug("Getting matches for scores with effect allele only")
    matches.append(
        match_variants(scorefile_no_oa, target, match_type="no_oa_ref").select(
            col_order
        )
    )
    matches.append(
        match_variants(scorefile_no_oa, target, match_type="no_oa_alt").select(
            col_order
        )
    )
    matches.append(
        match_variants(scorefile_no_oa, target, match_type="no_oa_ref_flip").select(
            col_order
        )
    )
    matches.append(
        match_variants(scorefile_no_oa, target, match_type="no_oa_alt_flip").select(
            col_order
        )
    )

    return matches


def match_variants(
    scorefile: pl.LazyFrame, target: pl.LazyFrame, match_type: str
) -> pl.LazyFrame:
    logger.debug(f"Matching strategy: {match_type}")
    match match_type:
        case "refalt":
            score_keys = ["chr_name", "chr_position", "effect_allele", "other_allele"]
            target_keys = ["#CHROM", "POS", "REF", "ALT"]
            effect_allele_column = "effect_allele"
        case "altref":
            score_keys = ["chr_name", "chr_position", "effect_allele", "other_allele"]
            target_keys = ["#CHROM", "POS", "ALT", "REF"]
            effect_allele_column = "effect_allele"
        case "refalt_flip":
            score_keys = [
                "chr_name",
                "chr_position",
                "effect_allele_FLIP",
                "other_allele_FLIP",
            ]
            target_keys = ["#CHROM", "POS", "REF", "ALT"]
            effect_allele_column = "effect_allele_FLIP"
        case "altref_flip":
            score_keys = [
                "chr_name",
                "chr_position",
                "effect_allele_FLIP",
                "other_allele_FLIP",
            ]
            target_keys = ["#CHROM", "POS", "ALT", "REF"]
            effect_allele_column = "effect_allele_FLIP"
        case "no_oa_ref":
            score_keys = ["chr_name", "chr_position", "effect_allele"]
            target_keys = ["#CHROM", "POS", "REF"]
            effect_allele_column = "effect_allele"
        case "no_oa_alt":
            score_keys = ["chr_name", "chr_position", "effect_allele"]
            target_keys = ["#CHROM", "POS", "ALT"]
            effect_allele_column = "effect_allele"
        case "no_oa_ref_flip":
            score_keys = ["chr_name", "chr_position", "effect_allele_FLIP"]
            target_keys = ["#CHROM", "POS", "REF"]
            effect_allele_column = "effect_allele_FLIP"
        case "no_oa_alt_flip":
            score_keys = ["chr_name", "chr_position", "effect_allele_FLIP"]
            target_keys = ["#CHROM", "POS", "ALT"]
            effect_allele_column = "effect_allele_FLIP"
        case _:
            logger.critical(f"Invalid match strategy: {match_type}")
            raise Exception

    missing_cols = ["REF", "ALT"]
    if match_type.startswith("no_oa"):
        if match_type.startswith("no_oa_ref"):
            missing_cols = ["REF"]
        else:
            missing_cols = ["ALT"]
    join_cols = ["ID"] + missing_cols
    return (
        scorefile.join(
            other=target, left_on=score_keys, right_on=target_keys, how="inner"
        )
        .with_columns(
            [
                pl.col("*"),
                pl.col(effect_allele_column).alias("matched_effect_allele"),
                pl.lit(match_type).alias("match_type"),
            ]
        )
        # note: get REF / ALT back after first join
        .join(target.select(join_cols), on="ID", how="inner")
    )



import logging

import polars as pl

logger = logging.getLogger(__name__)


def filter_target(df: pl.LazyFrame) -> pl.LazyFrame:
    """Remove variants that won't be matched against the scorefile

    Chromosomes 1 - 22, X, and Y with an efficient join. Remmove variants with missing identifiers also
    """
    logger.debug("Filtering target to include chromosomes 1 - 22, X, Y")
    chroms = [str(x) for x in list(range(1, 23)) + ["X", "Y"]]
    return df.filter((pl.col("#CHROM").is_in(chroms)) & (pl.col("ID") != "."))


def complement_valid_alleles(df: pl.LazyFrame, flip_cols: list[str]) -> pl.LazyFrame:
    """Improved function to complement alleles. Will only complement sequences that are valid DNA."""
    for col in flip_cols:
        logger.debug(f"Complementing column {col}")
        new_col = col + "_FLIP"
        df = df.with_columns(
            pl.when(pl.col(col).str.contains("^[ACGT]+$"))
            .then(
                pl.col(col)
                .str.replace_all("A", "V")
                .str.replace_all("T", "X")
                .str.replace_all("C", "Y")
                .str.replace_all("G", "Z")
                .str.replace_all("V", "T")
                .str.replace_all("X", "A")
                .str.replace_all("Y", "G")
                .str.replace_all("Z", "C")
            )
            .otherwise(pl.col(col))
            .alias(new_col)
        )
    return df


def annotate_multiallelic(df: pl.LazyFrame) -> pl.LazyFrame:
    """Identify variants that are multiallelic with a column flag"""
    # plink2 pvar multi-alleles are comma-separated
    df: pl.LazyFrame = df.with_columns(
        pl.when(pl.col("ALT").str.contains(","))
        .then(pl.lit(True))
        .otherwise(pl.lit(False))
        .alias("is_multiallelic")
    )

    if (
        df.select("is_multiallelic").unique().collect().get_column("is_multiallelic")
    ).any():
        logger.debug("Exploding dataframe to handle multiallelic variants")
        return df.with_columns(pl.col("ALT").str.split(by=",")).explode("ALT")
    else:
        logger.debug("No multiallelic variants detected")
        return df



import argparse
import atexit
import logging
import pathlib
import shutil
import sys
import tempfile
import textwrap

import polars as pl

from .. import (
    VariantFrame,
    ScoringFileFrame,
    match_variants,
    MatchResult,
    MatchResults,
)

from ._config import Config
from ._write import write_matches

logger = logging.getLogger(__name__)


def _exit(cleanup):
    if cleanup:
        if Config.TMPDIR and Config.TMPDIR.exists():
            shutil.rmtree(Config.TMPDIR)
        if Config.MATCHTMP and Config.MATCHTMP.exists():
            shutil.rmtree(Config.MATCHTMP)


def run_match():
    args = parse_args()

    if args.verbose:
        logging.getLogger("pgscatalog.core").setLevel(logging.DEBUG)
        logging.getLogger("pgscatalog.match").setLevel(logging.DEBUG)
        logger.setLevel(logging.DEBUG)
        logger.debug("Verbose logging enabled")

    Config.DATASET = args.dataset
    Config.CLEANUP = args.cleanup
    Config.OUTDIR = pathlib.Path(args.outdir)

    Config.SCOREFILE = pathlib.Path(args.scorefile)

    if not Config.OUTDIR.exists():
        raise FileNotFoundError(f"{Config.OUTDIR} does not exist")

    atexit.register(_exit, cleanup=Config.CLEANUP)

    if Config.CLEANUP:
        logger.info("--cleanup set (default), temporary files will be deleted")
    else:
        logger.info("--no-cleanup set, temporary files won't be deleted")

    Config.MATCHTMP = Config.OUTDIR / "matchtmp"
    Config.MATCHTMP.mkdir(exist_ok=False)
    Config.TMPDIR = Config.OUTDIR / "tmp"
    Config.TMPDIR.mkdir(exist_ok=False)
    Config.CHROM = args.chrom
    Config.MIN_OVERLAP = args.min_overlap

    if (n_target := len(args.target)) == 0:
        raise ValueError(f"{n_target=} must provide at least one target")

    Config.SPLIT = args.split
    Config.COMBINED = args.combined

    # parameters that control how the best match candidate is chosen
    # missing parameters will be set to defaults specified in matchlib
    Config.MATCH_PARAMS = {
        k: v for k in Config.MATCH_KWARGS if (v := getattr(args, k)) is not None
    }

    # start doing the work
    with ScoringFileFrame(
        path=Config.SCOREFILE,
        chrom=Config.CHROM,
        cleanup=Config.CLEANUP,
        tmpdir=Config.TMPDIR,
    ) as score_df, pl.StringCache():
        matchresults = []
        for target_f in args.target:
            ipc_path = get_match_candidates(
                target=target_f,
                score_df=score_df,
                dataset=Config.DATASET,
                tmpdir=Config.TMPDIR,
                chrom=Config.CHROM,
                cleanup=Config.CLEANUP,
            )
            # MatchResult is now backed by a stable-ish IPC file that will rarely get cleaned up
            matchresults.append(
                MatchResult.from_ipc(
                    matchresults_ipc_path=ipc_path, dataset=Config.DATASET
                )
            )

        if args.only_match:
            for i, match in enumerate(matchresults):
                ipc_path = pathlib.Path(match.ipc_path)
                out_path = Config.OUTDIR / f"{i}.ipc.zst"
                logger.info(f"Renaming {ipc_path} to {out_path}")
                ipc_path.rename(out_path)

            logger.warning("--only_match set, exiting with grace and aplomb")
            logger.warning(
                f"You'll need to combine match candidates using the Arrow IPC files in {Config.OUTDIR}"
            )
            return
        else:
            matchresults = MatchResults(*matchresults)
            _ = write_matches(matchresults=matchresults, score_df=score_df)
    logger.info("finished matching :)")


def get_match_candidates(target, score_df, chrom, dataset, **kwargs):
    # don't clean this up, useful for debugging
    _, fout = tempfile.mkstemp(dir=Config.MATCHTMP)  # file names don't matter
    variants = VariantFrame(path=target, chrom=chrom, dataset=dataset, **kwargs)
    with variants as target_df:
        _ = match_variants(
            score_df=score_df, target_df=target_df, target=variants
        ).collect(outfile=fout)

    return fout


def parse_args(args=None):
    parser = argparse.ArgumentParser(
        description=_description_text(),
        epilog=_epilog_text(),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "-d",
        "--dataset",
        dest="dataset",
        required=True,
        help="<Required> Label for target genomic dataset",
    )
    parser.add_argument(
        "-s",
        "--scorefiles",
        dest="scorefile",
        required=True,
        help="<Required> Combined scorefile path (output of read_scorefiles.py)",
    )
    parser.add_argument(
        "-t",
        "--target",
        dest="target",
        required=True,
        nargs="+",
        help="<Required> A list of paths of target genomic variants (.bim format)",
    )
    parser.add_argument(
        "-c",
        "--chrom",
        dest="chrom",
        required=False,
        type=str,
        help="<Optional> Set which chromosome is in the target variant file to speed up matching ",
    )
    parser.add_argument(
        "--only_match",
        dest="only_match",
        action="store_true",
        help="<Optional> Only match, then write intermediate files, don't make scoring files",
    )
    parser.add_argument(
        "--min_overlap",
        dest="min_overlap",
        required=False,
        type=float,
        help="<Optional> Minimum proportion of variants to match before error",
    )
    parser.add_argument(
        "-IDs",
        "--filter_IDs",
        dest="filter_IDs",
        help="<Optional> Path to file containing list of variant IDs that can be included in the final scorefile."
        "[useful for limiting scoring files to variants present in multiple datasets]",
    )
    parser.add_argument(
        "--outdir", dest="outdir", required=True, help="<Required> Output directory"
    )
    parser.add_argument(
        "--split",
        dest="split",
        default=False,
        action="store_true",
        help="<Optional> Split scorefile per chromosome?",
    )
    parser.add_argument(
        "--combined",
        dest="combined",
        default=False,
        action="store_true",
        help="<Optional> Write scorefiles in combined format?",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        help="<Optional> Extra logging information",
    )
    parser.add_argument(
        "--cleanup", dest="cleanup", default=True, action=argparse.BooleanOptionalAction
    )
    # variant matching arguments -------------------------------------------------------
    parser.add_argument(
        "--keep_ambiguous",
        dest="remove_ambiguous",
        action="store_false",
        help="""<Optional> Flag to force the program to keep variants with
                        ambiguous alleles, (e.g. A/T and G/C SNPs), which are normally
                        excluded (default: false). In this case the program proceeds
                        assuming that the genotype data is on the same strand as the
                        GWAS whose summary statistics were used to construct the score.
                                """,
    )
    parser.add_argument(
        "--keep_multiallelic",
        dest="remove_multiallelic",
        action="store_false",
        help="<Optional> Flag to allow matching to multiallelic variants (default: false).",
    )
    parser.add_argument(
        "--ignore_strand_flips",
        dest="skip_flip",
        action="store_true",
        help="""<Optional> Flag to not consider matched variants that may be reported 
                        on the opposite strand.  Default behaviour is to flip/complement unmatched variants and check if
                        they match.""",
    )
    parser.add_argument(
        "--keep_first_match",
        dest="keep_first_match",
        action="store_true",
        help="""<Optional> If multiple match candidates for a variant exist that can't be prioritised,
                         keep the first match candidate (default: drop all candidates)""",
    )
    return _check_args(parser.parse_args(args))


def _description_text() -> str:
    return textwrap.dedent(
        """\
    Match variants from a combined scoring file against a set of
    target genomes from the same fileset, and output scoring files
    compatible with the plink2 --score function.

    A combined scoring file is the output of the combine_scorefiles
    script. It has the following structure:

        | chr_name | chr_position | ... | accession |
        | -------- | ------------ | --- | --------- |
        | 1        | 1            | ... | PGS000802 |

    The combined scoring file is in long format, with one row per
    variant for each scoring file (accession). This structure is
    different to the PGS Catalog standard, because the long format
    makes matching faster and simpler.

    Target genomes can be in plink1 bim format or plink2 pvar
    format. Variant IDs should be unique so that they can be specified
    in the scoring file as: variant_id|effect_allele|[effect_weight column(s)...] 

    Only one set of target genomes should be matched at a time. Don't
    try to match target genomes from different plink filesets. Matching 
    against a set of chromosomes from the same fileset is OK (see --split). 
   """
    )


def _epilog_text() -> str:
    return textwrap.dedent(
        """\
    match_variants will output at least one scoring file in a
    format compatible with the plink2 --score function. This
    output might be split across different files to ensure each
    variant ID, effect allele, and effect type appears only once
    in each file. Output files have the pattern:

        {dataset}_{chromosome}_{effect_type}_{n}.scorefile.

    If multiple chromosomes are combined into a single file (i.e. not
    --split), then {chromosome} is replaced with 'ALL'. Once the
    scorefiles are used to calculate a score with plink2, the .sscore
    files will need to be aggregated to calculate a single polygenic
    score for each dataset, sample, and accession (scoring file). The
    PGS Catalog Calculator does this automatically.
    """
    )


def _check_args(args):
    if args.chrom is not None and not args.only_match:
        # filtering the scoring file will break overlap assumptions and calculations
        # e.g.:
        #   what if one chromosome matches well but another chromosome matches poorly?
        #   what if the poorly matching chromosome only has 5 variants to match?
        #
        # pgsc_calc uses global overlap % to decide if a score fails matching
        # --only_match skips overlap calculations (done in combine_matches instead)
        logger.critical("--chrom requires --only_match")
        sys.exit(1)
    if args.only_match and args.min_overlap is not None:
        # can't calculate min_overlap properly if just checking matches
        logger.critical("Invalid arguments: --only_match and --min_overlap (pick one!)")
        sys.exit(1)
    if not args.only_match and args.min_overlap is None:
        # need to calculate min_overlap before making scoring files
        logger.critical("Invalid arguments: set --min_overlap or --only_match")
        sys.exit(1)
    if args.split and args.only_match:
        # not writing scoring files, so split output doesn't make sense
        logger.critical("Invalid arguments: --only_match and --split (pick one!)")
        sys.exit(1)
    if any(
        [
            x in sys.argv
            for x in [
                "--keep_first_match",
                "--ignore_strand_flips",
                "--keep_multiallelic",
                "--keep_ambiguous",
            ]
        ]
    ):
        logger.warning(
            "Invalid arguments: --only_match and --keep_first_match, --ignore_strand_flips,"
            "keep_multiallelic, or keep_ambiguous"
        )
        logger.warning("Pass these arguments to combine_matches instead")

    if args.combined is False and args.split is False:
        logger.warning("No output format specified, writing to combined scoring file")
        args.combined = True

    return args


if __name__ == "__main__":
    run_match()



import collections.abc
import logging

import polars as pl

from pgscatalog.core import ZeroMatchesError, MatchRateError

from ._plinkframe import PlinkFrames
from ._match.label import label_matches
from ._match.filter import filter_scores
from ._match.log import make_logs, check_log_count, make_summary_log

logger = logging.getLogger(__name__)


class MatchResult:
    """Represents variants in a scoring file matched against variants in a target genome

    When matching a scoring file, it's normal for matches to be composed of many
    :class:`MatchResult` objects. This is common if the target genome is split to
    have one chromosome per scoring file, and the container class
    :class:`MatchResults` provides some helpful methods for working with split data.

    >>> from ._config import Config
    >>> from .variantframe import VariantFrame
    >>> from .scoringfileframe import ScoringFileFrame, match_variants
    >>> target_path = Config.ROOT_DIR.parent / "pgscatalog.core" / "tests" / "data" / "hapnest.bim"
    >>> target = VariantFrame(target_path, dataset="hapnest")
    >>> score_path = Config.ROOT_DIR.parent / "pgscatalog.core" / "tests" / "data" / "combined.txt.gz"
    >>> scorefile = ScoringFileFrame(score_path)

    A :class:`MatchResult` can be instantiated with the lazyframe output of the match_variants function:

    >>> with target as target_df, scorefile as score_df:
    ...     match_variants(score_df=score_df, target_df=target_df, target=target)  # doctest: +ELLIPSIS
    MatchResult(dataset=hapnest, matchresult=[<LazyFrame...], ipc_path=None, df=None)

    A :class:`MatchResult` can also be saved to and loaded from Arrow IPC files:

    >>> import tempfile
    >>> fout = tempfile.NamedTemporaryFile(delete=False)
    >>> with target as target_df, scorefile as score_df:
    ...     results = match_variants(score_df=score_df, target_df=target_df, target=target)
    ...     _ = results.collect(outfile=fout.name)
    >>> x = MatchResult.from_ipc(fout.name, dataset="hapnest")
    >>> x  # doctest: +ELLIPSIS
    MatchResult(dataset=hapnest, matchresult=None, ipc_path=..., df=<LazyFrame...>)
    """

    def __init__(self, dataset, matchresult=None, ipc_path=None):
        # a dataset ID
        self.dataset = dataset
        self.ipc_path = ipc_path
        # df: the dataframe backed by stable arrow files at ipc_path
        # after collect() is written to a file
        self.df = None

        # a lazyframe of match results, backed by arrow files that may get cleaned up
        self._matchresult = matchresult

        match (self.ipc_path, matchresult):
            case (_, None):
                # init from results already written to a file
                self.df = pl.scan_ipc(self.ipc_path)
            case (None, _):
                # init from a lazy frame result from get_all_matches()
                # needs to be collected() later and optionally written to an IPC file
                pass
            case (None, None) | (_, _):
                raise ValueError

    def collect(self, outfile=None):
        """Compute match results and optionally save to file"""
        if self._matchresult is None:
            raise ValueError("Can't collect, missing matchresult")

        if outfile is not None:
            pl.concat(pl.collect_all(self._matchresult)).write_ipc(outfile)
            self.df = pl.scan_ipc(outfile)
        else:
            # call .lazy() to prevent eager queries later
            self.df = pl.concat(pl.collect_all(self._matchresult)).lazy()

        return self.df

    @classmethod
    def from_ipc(cls, matchresults_ipc_path, dataset):
        """Create an instance from an Arrow IPC file"""
        return cls(
            ipc_path=matchresults_ipc_path,
            dataset=dataset,
        )

    def __repr__(self):
        return f"{type(self).__name__}(dataset={self.dataset}, matchresult={self._matchresult!r}, ipc_path={self.ipc_path}, df={self.df!r})"


class MatchResults(collections.abc.Sequence):
    """
    Container for :class:`MatchResult`

    Useful for making matching logs and writing scoring files ready to be used by ``plink2 --score``

    >>> import tempfile, os, glob, pathlib
    >>> from ._config import Config
    >>> from .variantframe import VariantFrame
    >>> from .scoringfileframe import ScoringFileFrame, match_variants
    >>> fout = tempfile.NamedTemporaryFile(delete=False)
    >>> target_path = Config.ROOT_DIR / "tests" / "data" / "good_match.pvar"
    >>> score_path =  Config.ROOT_DIR / "tests" / "data" / "good_match_scorefile.txt"
    >>> target = VariantFrame(target_path, dataset="goodmatch")
    >>> scorefile = ScoringFileFrame(score_path)
    >>> foutdir, splitfoutdir = tempfile.mkdtemp(), tempfile.mkdtemp()

    Using a context manager is really important to prepare :class:`ScoringFileFrame` and :class:`VariantFrame` data frames:

    >>> with target as target_df, scorefile as score_df:
    ...     results = match_variants(score_df=score_df, target_df=target_df, target=target)
    ...     _ = results.collect(outfile=fout.name)

    These data frames are transparently backed by Arrow IPC files on disk.

    >>> with scorefile as score_df:
    ...     x = MatchResult.from_ipc(fout.name, dataset="goodmatch")
    ...     _ = MatchResults(x).write_scorefiles(directory=foutdir, score_df=score_df)
    ...     _ = MatchResults(x).write_scorefiles(directory=splitfoutdir, split=True, score_df=score_df)
    >>> MatchResults(x)  # doctest: +ELLIPSIS
    MatchResults([MatchResult(dataset=goodmatch, matchresult=None, ipc_path=...])

    By default, scoring files are written with multiple chromosomes per file:

    >>> combined_paths = sorted(glob.glob(foutdir + "/*ALL*"), key=lambda x: pathlib.Path(x).stem)
    >>> combined_paths # doctest: +ELLIPSIS
    ['.../goodmatch_ALL_additive_0.scorefile.gz', '.../goodmatch_ALL_dominant_0.scorefile.gz', '.../goodmatch_ALL_recessive_0.scorefile.gz']
    >>> assert len(combined_paths) == 3

    Scoring files can be split. The input scoring file contains 20 unique
    chromosomes, with one additive + dominant effect file (but one chromosome didn't match well):

    >>> scorefiles = sorted(os.listdir(splitfoutdir))
    >>> scorefiles # doctest: +ELLIPSIS
    ['goodmatch_10_additive_0.scorefile.gz', 'goodmatch_11_additive_0.scorefile.gz', ...]
    >>> sum("dominant" in f for f in scorefiles)
    1
    >>> sum("recessive" in f for f in scorefiles)
    1
    >>> sum("additive" in f for f in scorefiles)
    19
    >>> assert len(scorefiles) == 21

    An important part of matching variants is reporting a log to see how well you're reproducing a PGS in the new target genomes:

    >>> with scorefile as score_df:
    ...     MatchResults(x).full_variant_log(score_df).fetch()  # +ELLIPSIS
    shape: (155, 23)
    ┌────────┬───────────┬──────────┬────────────┬───┬────────────┬───────────┬────────────┬───────────┐
    │ row_nr ┆ accession ┆ chr_name ┆ chr_positi ┆ … ┆ duplicate_ ┆ match_IDs ┆ match_stat ┆ dataset   │
    │ ---    ┆ ---       ┆ ---      ┆ on         ┆   ┆ ID         ┆ ---       ┆ us         ┆ ---       │
    │ u64    ┆ cat       ┆ cat      ┆ ---        ┆   ┆ ---        ┆ str       ┆ ---        ┆ cat       │
    │        ┆           ┆          ┆ u64        ┆   ┆ bool       ┆           ┆ cat        ┆           │
    ╞════════╪═══════════╪══════════╪════════════╪═══╪════════════╪═══════════╪════════════╪═══════════╡
    │ 0      ┆ PGS000002 ┆ 11       ┆ 69331418   ┆ … ┆ null       ┆ null      ┆ unmatched  ┆ goodmatch │
    │ 1      ┆ PGS000002 ┆ 11       ┆ 69379161   ┆ … ┆ false      ┆ NA        ┆ matched    ┆ goodmatch │
    │ 2      ┆ PGS000002 ┆ 11       ┆ 69331642   ┆ … ┆ null       ┆ null      ┆ unmatched  ┆ goodmatch │
    │ 3      ┆ PGS000002 ┆ 5        ┆ 1282319    ┆ … ┆ false      ┆ NA        ┆ matched    ┆ goodmatch │
    │ 4      ┆ PGS000002 ┆ 5        ┆ 1279790    ┆ … ┆ false      ┆ NA        ┆ matched    ┆ goodmatch │
    │ …      ┆ …         ┆ …        ┆ …          ┆ … ┆ …          ┆ …         ┆ …          ┆ …         │
    │ 72     ┆ PGS000001 ┆ 22       ┆ 40876234   ┆ … ┆ false      ┆ NA        ┆ matched    ┆ goodmatch │
    │ 73     ┆ PGS000001 ┆ 1        ┆ 204518842  ┆ … ┆ false      ┆ NA        ┆ matched    ┆ goodmatch │
    │ 74     ┆ PGS000001 ┆ 1        ┆ 202187176  ┆ … ┆ false      ┆ NA        ┆ matched    ┆ goodmatch │
    │ 75     ┆ PGS000001 ┆ 2        ┆ 19320803   ┆ … ┆ false      ┆ NA        ┆ matched    ┆ goodmatch │
    │ 76     ┆ PGS000001 ┆ 16       ┆ 53855291   ┆ … ┆ null       ┆ null      ┆ unmatched  ┆ goodmatch │
    └────────┴───────────┴──────────┴────────────┴───┴────────────┴───────────┴────────────┴───────────┘
    """

    def __init__(self, *elements):
        self._elements = list(elements)

        if len(datasets := set(x.dataset for x in elements)) > 1:
            raise ValueError(
                f"MatchResult elements must have same dataset, but found: {datasets!r}"
            )

        self.dataset = self._elements[0].dataset
        # a df composed of all match result elements
        self.df = pl.scan_ipc(x.ipc_path for x in self._elements)
        if self.df.select("row_nr").collect().is_empty():
            raise ZeroMatchesError("No match candidates found for any scoring files")

        # a table containing up to one row per variant (the best possible match)
        self.filter_summary = None
        # a summary log containing match rates for variants
        self.summary_log = None
        # have match candidates in df been labelled?
        self._labelled = False
        # have match candidates in df been filtered?
        self._filtered = False
        # does the input scoring file count match the variant log count?
        self._log_OK = None

    def __len__(self):
        return len(self._elements)

    def __getitem__(self, item):
        return self._elements[item]

    def __repr__(self):
        return f"{type(self).__name__}({self._elements!r})"

    def label(
        self,
        keep_first_match=False,
        remove_ambiguous=True,
        skip_flip=False,
        remove_multiallelic=True,
        filter_IDs=None,
    ):
        """Label match candidates according to matching parameters

        kwargs control labelling parameters:

        * ``keep_first_match``: if best match candidates are tied, keep the first? (default: ```False``, drop all candidates for this variant)
        * ``remove_ambiguous``: Remove ambiguous alleles? (default: ``True``)
        * ``skip_flip``: Consider matched variants that may be reported on the opposite strand (default: ``False``)
        * ``remove_multiallelic`` remove multiallelic variants before matching (default: ``True``)
        * ``filter_IDs``: constrain variants to this list of IDs (default, don't constrain)
        """
        df = self.df.pipe(
            label_matches,
            keep_first_match=keep_first_match,
            remove_ambiguous=remove_ambiguous,
            skip_flip=skip_flip,
            remove_multiallelic=remove_multiallelic,
            filter_IDs=filter_IDs,
        )
        self._labelled = True
        self.df = df
        return self.df

    def filter(self, score_df, min_overlap=0.75, **kwargs):
        """Filter match candidates after labelling according to user parameters"""
        if not self._labelled:
            self.df = self.label(**kwargs)

        df, filter_summary = filter_scores(
            scorefile=score_df,
            matches=self.df,
            min_overlap=min_overlap,
            dataset=self.dataset,
        )
        self.filter_summary = filter_summary
        self.df = df
        self._filtered = True
        return self.df

    def write_scorefiles(
        self, directory, score_df, split=False, min_overlap=0.75, **kwargs
    ):
        """Write matches to a set of files ready for ``plink2 --score``

        Does some helpful stuff:

        * Labels match candidates
        * Filters match candidates based on labels and user configuration
        * Calculates match rates to see how well the PGS reproduces in the new target genomes
        * Generates a filtered variant log containing the best match candidate
        * Checks if the number of variants in the summary log matches the input scoring file
        * Sets up parallel score calculation (pivots data to wide column format)
        * Writes scores to a directory, splitting based on chromosome and effect type
        """
        if not self._labelled:
            _ = self.label(**kwargs)  # self.df gets updated

        if not self._filtered:
            # score_df = original scoring file variants
            _ = self.filter(score_df=score_df, min_overlap=min_overlap)

        # a summary log contains up to one variant (the best match) for each variant
        # in the scoring file
        self.summary_log = make_summary_log(
            match_candidates=self.df,
            dataset=self.dataset,
            filter_summary=self.filter_summary.lazy(),
            scorefile=score_df,
        )

        # double check log count vs scoring file variant count
        self._log_OK = check_log_count(scorefile=score_df, summary_log=self.summary_log)

        # will be empty if no scores pass match threshold, so nothing gets written
        plink = PlinkFrames.from_matchresult(self.df)
        outfs = []
        for frame in plink:
            f = frame.write(directory=directory, split=split, dataset=self.dataset)
            outfs.append(f)

        # collect after joining in check_log_count (can't join df and lazy df)
        self.summary_log = self.summary_log.collect()

        # error at the end, to allow logs to be generated
        for x in self.filter_summary.iter_rows():
            try:
                if not x[1]:
                    raise MatchRateError("Fails matching")
            except MatchRateError:
                # we reported the exception, but it's fine
                logger.warning(f"Score {x[0]} matching failed with match rate {x[2]}")
                continue
            else:
                logger.info(f"Score {x[0]} matching passed with match rate {x[2]}")

        # unless everything is bad!
        if not any(x[1] for x in self.filter_summary.iter_rows()):
            raise ZeroMatchesError(
                f"All scores fail to meet match threshold {min_overlap}"
            )

        return outfs

    def full_variant_log(self, score_df, **kwargs):
        """Generate a log for each variant in a scoring file

        Multiple match candidates may exist for each variant in the original file.
        Describe each variant (one variant per row) with match metadata
        """
        if not self._labelled:
            self.df = self.label(**kwargs)
            self._labelled = True

        if not self._filtered:
            self.df, self.filter_summary = filter_scores(
                scorefile=score_df,
                matches=self.df,
                min_overlap=kwargs.get("min_overlap", 0.75),
                dataset=self.dataset,
            )
            self._filtered = True

        return make_logs(
            scorefile=score_df, dataset=self.dataset, match_candidates=self.df
        )



"""This module contains the PlinkScoreFiles class, which represents one or more
scoring files ready to be used with plink2 --score"""

import collections.abc
import gzip
import io
import itertools
import logging
import pathlib

import polars as pl

from pgscatalog.core import chrom_keyfunc, effect_type_keyfunc

logger = logging.getLogger(__name__)


class PlinkScoreFiles(collections.abc.Sequence):
    """Represents a sequence of scoring files written by :class:`MatchResults`"""

    def __init__(self, *elements):
        self._elements = [pathlib.Path(x) for x in sorted(list(elements))]

    def __len__(self):
        return len(self._elements)

    def __getitem__(self, item):
        return self._elements[item]

    def __repr__(self):
        return f"{type(self).__name__}({self._elements!r})"

    def merge(self, directory):
        """Merge scoring files without recomputing matches

        Assumes a standard file naming system was used: ``dataset_chrom_effecttype_n``

        >>> import tempfile, os, glob
        >>> from ._config import Config
        >>> from .variantframe import VariantFrame
        >>> from .scoringfileframe import ScoringFileFrame, match_variants
        >>> from .matchresult import MatchResult, MatchResults
        >>> fout = tempfile.NamedTemporaryFile(delete=False)
        >>> target_path = Config.ROOT_DIR / "tests" / "data" / "good_match.pvar"
        >>> score_path =  Config.ROOT_DIR / "tests" / "data" / "good_match_scorefile.txt"
        >>> target = VariantFrame(target_path, dataset="goodmatch")
        >>> scorefile = ScoringFileFrame(score_path)
        >>> with target as target_df, scorefile as score_df:
        ...     results = match_variants(score_df=score_df, target_df=target_df, target=target)
        ...     _ = results.collect(outfile=fout.name)
        >>> x = MatchResult.from_ipc(fout.name, dataset="goodmatch")
        >>> foutdir = tempfile.mkdtemp()
        >>> with scorefile as score_df:
        ...     _ = MatchResults(x).write_scorefiles(directory=foutdir, split=True, score_df=score_df)  # doctest: +ELLIPSIS
        >>> plink_files = (pathlib.Path(foutdir) / x for x in os.listdir(foutdir))
        >>> psf = PlinkScoreFiles(*plink_files)
        >>> psf  # doctest: +ELLIPSIS
        PlinkScoreFiles([PosixPath('.../goodmatch_1_additive_0.scorefile.gz'), ...])
        >>> psf.merge(foutdir)
        >>> combined_paths = sorted(glob.glob(foutdir + "/*ALL*"),  key=lambda x: pathlib.Path(x).stem)
        >>> len(combined_paths)
        3
        >>> combined_paths # doctest: +ELLIPSIS
        ['.../goodmatch_ALL_additive_0.scorefile.gz', '.../goodmatch_ALL_dominant_0.scorefile.gz', '.../goodmatch_ALL_recessive_0.scorefile.gz']

        """
        dataset = self._elements[0].stem.split("_")[0]
        for x in self._elements:
            if dataset not in x.stem:
                raise ValueError(f"Invalid dataset: {dataset} and {x.stem}")

        sorted_paths = sorted(self._elements, key=effect_type_keyfunc())

        for k, g in itertools.groupby(sorted_paths, key=effect_type_keyfunc()):
            logger.info(f"Writing combined scoring file {k}")

            # tidy up keys to create the output file name
            # keyfunc returns:
            # (('additive',), ('', 0.0, 'scorefile'))
            # need: [additive, 0]
            keys = (x for x in (itertools.chain(*k)) if x != "")
            keys = list(str(int(x)) if isinstance(x, float) else x for x in keys)
            keys.pop()  # drop scorefile

            # multi-chrom -> ALL
            fout = "_".join([dataset, "ALL", *keys]) + ".scorefile.gz"
            paths = sorted(list(g), key=chrom_keyfunc())

            # infer_schema_length: read all columns as utf8 to simplify joining
            dfs = (pl.read_csv(x, separator="\t", infer_schema_length=0) for x in paths)
            # diagonal concat is important to handle different column sets across dfs
            df = pl.concat(dfs, how="diagonal").fill_null(value="0")
            logger.info("Score files combined successfully")

            with gzip.open(
                pathlib.Path(directory) / fout, "wb", compresslevel=6
            ) as gcsv:
                logger.info(f"Writing out to {fout}")
                outf = io.TextIOWrapper(gcsv)
                df.write_csv(outf, separator="\t")



import argparse
import logging
import pathlib

from xopen import xopen
import csv
import textwrap
import heapq
import tempfile


logger = logging.getLogger(__name__)


def run_intersect():
    args = parse_args()
    outdir = pathlib.Path(args.outdir)

    if not outdir.is_dir():
        raise NotADirectoryError

    tmpdir = tempfile.mkdtemp()

    if args.verbose:
        logging.getLogger("pgscatalog.core").setLevel(logging.DEBUG)
        logging.getLogger("pgscatalog.match").setLevel(logging.DEBUG)
        logger.setLevel(logging.DEBUG)
        logger.debug("Verbose logging enabled")

    # Process & sort reference variants
    count_var_r = 0
    o_tmp_r = []
    logger.info("Reading REFERENCE variants: {}".format(args.reference))
    ref_heap = []
    ref_pvar = read_var_general(args.reference, chrom=args.filter_chrom)
    for v in ref_pvar:
        count_var_r += 1
        ALTs = v["ALT"].split(",")
        IS_MA_REF = len(ALTs) > 1
        for i, ALT in enumerate(ALTs):
            if v["REF"] < ALT:
                key = "{}:{}:{}:{}".format(v["#CHROM"], v["POS"], v["REF"], ALT)
            else:
                key = "{}:{}:{}:{}".format(v["#CHROM"], v["POS"], ALT, v["REF"])

            IS_INDEL = (len(v["REF"]) > 1) | (len(ALT) > 1)
            STRANDAMB = v["REF"] == allele_complement(ALT)
            ref_heap.append(
                ([key, v["ID"], v["REF"]], [IS_INDEL, STRANDAMB, IS_MA_REF])
            )

        if count_var_r % 500000 == 0:
            heapq.heapify(ref_heap)
            tmppath = tempfile.NamedTemporaryFile(dir=tmpdir, delete=False)
            with open(tmppath.name, "wt") as outf:
                o_tmp_r.append(tmppath.name)
                outf.write(
                    "CHR:POS:A0:A1\tID_REF\tREF_REF\tIS_INDEL\tSTRANDAMB\tIS_MA_REF\n"
                )
                for i in range(len(ref_heap)):
                    popped = heapq.heappop(ref_heap)
                    outf.write(
                        "\t".join([str(x) for x in popped[0] + popped[1]]) + "\n"
                    )
            ref_heap = []
            logger.info("Processed {} REFERENCE variants".format(count_var_r))

    if len(ref_heap) > 0:
        heapq.heapify(ref_heap)
        tmppath = tempfile.NamedTemporaryFile(dir=tmpdir, delete=False)
        with open(tmppath.name, "wt") as outf:
            o_tmp_r.append(tmppath.name)
            outf.write(
                "CHR:POS:A0:A1\tID_REF\tREF_REF\tIS_INDEL\tSTRANDAMB\tIS_MA_REF\n"
            )
            for i in range(len(ref_heap)):
                popped = heapq.heappop(ref_heap)
                outf.write("\t".join([str(x) for x in popped[0] + popped[1]]) + "\n")
        del ref_heap
        logger.info("Processed {} REFERENCE variants".format(count_var_r))

    logger.info("Outputting REFERNCE variants -> reference_variants.txt.gz")
    with xopen(outdir / "reference_variants.txt.gz", "wt") as outf:
        outf.write("CHR:POS:A0:A1\tID_REF\tREF_REF\tIS_INDEL\tSTRANDAMB\tIS_MA_REF\n")
        for v in heapq.merge(
            *[read_var_general(x) for x in o_tmp_r],
            key=lambda v: (v["CHR:POS:A0:A1"], v["ID_REF"], v["REF_REF"]),
        ):
            outf.write("\t".join(v.values()) + "\n")

    # Process & sort target variants
    count_var_t = 0
    o_tmp_t = []
    target_heap = []
    for path in args.target:
        logger.info("Reading TARGET variants: {}".format(path))
        pvar = read_var_general(
            path, chrom=None
        )  # essential not to filter target (messes up common line indexing)

        loc_afreq = path.replace(".pvar.zst", ".afreq.gz")
        afreq = read_var_general(
            loc_afreq, chrom=None
        )  # essential not to filter target (messes up common line indexing)

        loc_vmiss = path.replace(".pvar.zst", ".vmiss.gz")
        vmiss = read_var_general(
            loc_vmiss, chrom=None
        )  # essential not to filter target (messes up common line indexing)

        for v, freq, miss in zip(pvar, afreq, vmiss):
            if all([v["ID"], freq["ID"], miss["#ID"]]) is False:
                raise ValueError("TARGET variant files are not sorted")
            count_var_t += 1
            ALTs = v["ALT"].split(",")
            ALT_FREQS = [float(x) for x in freq["ALT_FREQS"].split(",")]
            F_MISS_DOSAGE = miss["F_MISS_DOSAGE"]
            IS_MA_TARGET = len(ALTs) > 1
            for i, ALT in enumerate(ALTs):
                if v["REF"] < ALT:
                    key = "{}:{}:{}:{}".format(v["#CHROM"], v["POS"], v["REF"], ALT)
                else:
                    key = "{}:{}:{}:{}".format(v["#CHROM"], v["POS"], ALT, v["REF"])
                target_heap.append(
                    (
                        [key, v["ID"], v["REF"]],
                        [IS_MA_TARGET, ALT_FREQS[i], F_MISS_DOSAGE],
                    )
                )

            if count_var_t % 500000 == 0:
                heapq.heapify(target_heap)
                tmppath = tempfile.NamedTemporaryFile(dir=tmpdir, delete=False)
                with open(tmppath.name, "wt") as outf:
                    o_tmp_t.append(tmppath.name)
                    outf.write(
                        "CHR:POS:A0:A1\tID_TARGET\tREF_TARGET\tIS_MA_TARGET\tAAF\tF_MISS_DOSAGE\n"
                    )
                    for i in range(len(target_heap)):
                        popped = heapq.heappop(target_heap)
                        outf.write(
                            "\t".join([str(x) for x in popped[0] + popped[1]]) + "\n"
                        )
                target_heap = []
                logger.info("Processed {} TARGET variants".format(count_var_t))

    if len(target_heap) > 0:
        heapq.heapify(target_heap)
        tmppath = tempfile.NamedTemporaryFile(dir=tmpdir, delete=False)
        with open(tmppath.name, "wt") as outf:
            o_tmp_t.append(tmppath.name)
            outf.write(
                "CHR:POS:A0:A1\tID_TARGET\tREF_TARGET\tIS_MA_TARGET\tAAF\tF_MISS_DOSAGE\n"
            )
            for i in range(len(target_heap)):
                popped = heapq.heappop(target_heap)
                outf.write("\t".join([str(x) for x in popped[0] + popped[1]]) + "\n")
        del target_heap
        logger.info("Processed {} TARGET variants".format(count_var_t))

    logger.info("Outputting TARGET variants -> target_variants.txt.gz")
    with xopen(outdir / "target_variants.txt.gz", "wt") as outf:
        outf.write(
            "CHR:POS:A0:A1\tID_TARGET\tREF_TARGET\tIS_MA_TARGET\tAAF\tF_MISS_DOSAGE\n"
        )
        for v in heapq.merge(
            *[read_var_general(x) for x in o_tmp_t],
            key=lambda v: (v["CHR:POS:A0:A1"], v["ID_TARGET"], v["REF_TARGET"]),
        ):
            outf.write("\t".join(v.values()) + "\n")

    # Merge matched variants on sorted files
    logger.info("Joining & outputting matched variants -> matched_variants.txt.gz")
    n_matched = 0
    n_PCA_ELIGIBLE = 0
    with xopen(outdir / "matched_variants.txt.gz", "w") as csvfile:
        for vmatch in sorted_join_variants(
            outdir / "reference_variants.txt.gz", outdir / "target_variants.txt.gz"
        ):
            n_matched += 1
            vmatch["SAME_REF"] = vmatch["REF_REF"] == vmatch["REF_REF"]

            # Define variant's eligibility for PCA
            # From original implementation: ((IS_MA_REF == FALSE) && (IS_MA_TARGET == FALSE)) && (((IS_INDEL == FALSE) && (STRANDAMB == FALSE)) || ((IS_INDEL == TRUE) && (SAME_REF == TRUE)))
            PCA_ELIGIBLE = (
                (vmatch["IS_MA_REF"] == "False") and (vmatch["IS_MA_TARGET"] == "False")
            ) and (
                ((vmatch["IS_INDEL"] == "False") and (vmatch["STRANDAMB"] == "False"))
                or ((vmatch["IS_INDEL"] == "True") and (vmatch["SAME_REF"] == "True"))
            )

            PCA_ELIGIBLE = (
                PCA_ELIGIBLE
                and (aaf2maf(float(vmatch["AAF"])) > args.maf_filter)
                and (float(vmatch["F_MISS_DOSAGE"]) < args.vmiss_filter)
            )
            vmatch["PCA_ELIGIBLE"] = PCA_ELIGIBLE
            if PCA_ELIGIBLE is True:
                n_PCA_ELIGIBLE += 1

            if n_matched == 1:
                writer = csv.DictWriter(
                    csvfile, fieldnames=vmatch.keys(), delimiter="\t"
                )
                writer.writeheader()
            writer.writerow(vmatch)
    logger.info(
        "{}/{} ({:.2f}%) of TARGET variants matched the REFERENCE data".format(
            n_matched, count_var_t, 100 * n_matched / count_var_t
        )
    )
    logger.info(
        "{}/{} ({:.2f}%) of matched variants are eligible for PCA".format(
            n_PCA_ELIGIBLE, n_matched, 100 * n_PCA_ELIGIBLE / n_matched
        )
    )

    # Output counts
    logger.info(
        "Outputting variant counts -> intersect_counts_{}.txt".format(args.filter_chrom)
    )
    with open(
        outdir / "intersect_counts_{}.txt".format(args.filter_chrom), "w"
    ) as outf:
        outf.write("\n".join(map(str, [count_var_t, count_var_r, n_matched])))


def read_var_general(path, chrom=None):
    """
    General function for reading variant files from plink2 outputs
    :param path: path to variant file
    :param chrom: filter to specific chromosome
    :return: row of a df as a dict
    """
    with xopen(path, "rt") as f:
        if "bim" in pathlib.Path(path).name:
            reader = csv.reader(f, delimiter="\t")
            # yes, A1/A2 in bim isn't ref/alt
            fields = ["#CHROM", "ID", "pos_cm", "POS", "REF", "ALT"]
            if (chrom is None) or (chrom == "ALL"):
                for row in reader:
                    yield dict(zip(fields, row, strict=True))
            else:
                for row in reader:
                    row = dict(zip(fields, row, strict=True))
                    if row["#CHROM"] == chrom:
                        yield row
        else:
            fields = None
            if (chrom is None) or (chrom == "ALL"):
                for row in f:
                    if row.startswith("##"):
                        continue
                    else:
                        row = row.strip().split("\t")
                        if fields is None:
                            fields = row
                        else:
                            yield dict(zip(fields, row, strict=True))
            else:
                for row in f:
                    if row.startswith("##"):
                        continue
                    else:
                        row = row.strip().split("\t")
                        if fields is None:
                            fields = row
                        else:
                            row = dict(zip(fields, row, strict=True))
                            if row["#CHROM"] == chrom:
                                yield row


def sorted_join_variants(path_ref, path_target):
    f1_iter = read_var_general(path_ref)
    f2_iter = read_var_general(path_target)

    prev_key1 = None  # Initialize previous key for file 1
    prev_key2 = None  # Initialize previous key for file 2

    line1 = next(f1_iter, None)
    line2 = next(f2_iter, None)

    while line1 is not None and line2 is not None:
        key1 = line1["CHR:POS:A0:A1"]
        key2 = line2["CHR:POS:A0:A1"]

        # Check if lines are sorted by the key for each file
        if prev_key1 is not None and key1 < prev_key1:
            raise ValueError("REFERENCE keys are not sorted")
        if prev_key2 is not None and key2 < prev_key2:
            raise ValueError("TARGET keys are not sorted")

        if key1 == key2:
            line1.update(line2)
            yield line1
            prev_key1 = key1  # Update previous key for file 1
            prev_key2 = key2  # Update previous key for file 2

            line1 = next(f1_iter, None)
            line2 = next(f2_iter, None)
        elif key1 < key2:
            prev_key1 = key1  # Update previous key for file 1
            line1 = next(f1_iter, None)
        else:
            prev_key2 = key2  # Update previous key for file 2
            line2 = next(f2_iter, None)


def allele_complement(s):
    """
    Complement alleles
    :param s: allele to be complemented
    :return: complement
    """
    return (
        s.replace("A", "V")
        .replace("T", "X")
        .replace("C", "Y")
        .replace("G", "Z")
        .replace("V", "T")
        .replace("X", "A")
        .replace("Y", "G")
        .replace("Z", "C")
    )


def aaf2maf(aaf):
    """
    Convert alternative allele frequency (AAF) to minor allele frequency (MAF)
    :param aaf: alternative allele frequency
    :return: minor allele frequency (MAF)
    """
    if aaf > 0.5:
        return 1 - aaf
    else:
        return aaf


def parse_args(args=None):
    parser = argparse.ArgumentParser(
        description=_description_text(),
        epilog=_epilog_text(),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "-r",
        "--reference",
        dest="reference",
        required=True,
        help="path/to/REFERENCE/pvar",
    )
    parser.add_argument(
        "-t",
        "--target",
        dest="target",
        required=True,
        nargs="+",
        help="<Required> A list of paths of target genomic variants (.bim/pvar format). The .afreq and .vmiss files are also required for these files.",
    )
    parser.add_argument(
        "-c",
        "--chrom",
        dest="filter_chrom",
        required=False,
        help="Whether to limit matches to specific chromosome of the reference",
    )
    parser.add_argument(
        "--maf_target",
        dest="maf_filter",
        default=0.05,
        type=float,
        required=False,
        help="Filter: Minimum minor Allele Frequency for PCA eligibility",
    )
    parser.add_argument(
        "--geno_miss",
        dest="vmiss_filter",
        default=0.1,
        type=float,
        required=False,
        help="Filter: Maximum Genotype missingness for PCA eligibility",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        help="<Optional> Extra logging information",
    )
    parser.add_argument(
        "--outdir", dest="outdir", required=True, help="<Required> Output directory"
    )
    return parser.parse_args(args)


def _description_text() -> str:
    return textwrap.dedent(
        """\
    Program to find matched variants (same strand) between a set of reference and target data .pvar/bim files. This 
    also uses the .afreq and .vmiss files to evaluate whether the variants in the TARGET are suitable for inclusion in
    a PCA analysis, while also filtering on strand ambiguity and multi-allelic/INDEL status. 
   """
    )


def _epilog_text() -> str:
    return textwrap.dedent("""""")


if __name__ == "__main__":
    run_intersect()



import logging
import os
import shutil

import polars as pl

from pgscatalog.core import TargetVariants

from ._arrow import loose
from ._match.preprocess import filter_target, annotate_multiallelic

logger = logging.getLogger(__name__)


class VariantFrame:
    """Similar to :class:`pgscatalog.core.TargetVariants`, but backed by the polars dataframe library

    Fast, supports more complicated things, but requires more resources.

    The context manager returns a polars LazyFrame:

    >>> from ._config import Config
    >>> path = Config.ROOT_DIR.parent / "pgscatalog.core" / "tests" / "data" / "hapnest.bim"
    >>> x = VariantFrame(path, dataset="hapnest")
    >>> with x as df:
    ...     df.collect().shape
    (101, 6)
    >>> x  # doctest: +ELLIPSIS
    VariantFrame(path='.../hapnest.bim', dataset='hapnest', chrom=None, cleanup=True, tmpdir=None)

    The :class:`VariantFrame` contains a :class:`pgscatalog.core.TargetVariants` object:

    >>> x.variants  # doctest: +ELLIPSIS
    TargetVariants(path='.../hapnest.bim')
    """

    def __init__(self, path, dataset, chrom=None, cleanup=True, tmpdir=None):
        self.variants = TargetVariants(path)
        # variant frames need an ID column (dataset)
        self.dataset = dataset
        self.chrom = chrom
        self._cleanup = cleanup
        self._tmpdir = tmpdir
        self._loosed = False
        self.arrowpaths = None

    def __enter__(self):
        # set up global string cache for categorical variables
        pl.enable_string_cache()

        if not self._loosed:
            logger.debug(f"Converting {self!r} to feather format")
            # convert to arrow files
            self.arrowpaths = loose(self.variants, tmpdir=self._tmpdir)
            self._loosed = True
            logger.debug(f"{self!r} feather conversion complete")

        target_df = (
            pl.scan_ipc(self.arrowpaths)
            .pipe(filter_target)
            .pipe(annotate_multiallelic)
            .with_columns(
                [
                    pl.col("#CHROM").cast(pl.Categorical),
                    pl.col("REF").cast(pl.Categorical),
                    pl.col("ALT").cast(pl.Categorical),
                ]
            )
        )

        return target_df

    def __exit__(self, *args, **kwargs):
        if self._cleanup:
            [os.unlink(x) for x in self.arrowpaths]
            self._loosed = False
        pl.disable_string_cache()

    def __repr__(self):
        return (
            f"{type(self).__name__}(path={repr(self.variants.path)}, dataset={repr(self.dataset)}, "
            f"chrom={repr(self.chrom)}, cleanup={repr(self._cleanup)}, "
            f"tmpdir={repr(self._tmpdir)})"
        )

    def save_ipc(self, destination):
        """Save the dataframe prepared by the context manager to an Arrow IPC file

        Useful because the context manager will clean up the IPC files while exiting.

        This method allows data to be persisted."""
        if not self._loosed:
            raise ValueError(
                "Can't save IPC because it doesn't exist."
                "Try calling inside a with: statement"
            )

        shutil.copyfile(self.arrowpaths, destination)



"""This module contains a simple class that stores CLI configuration as class attributes"""


class Config:
    # what's the label for the target variants data?
    DATASET = None
    # where do scorefiles get written to?
    OUTDIR = None
    # where do we keep stable-ish arrow files? (matchresults)
    MATCHTMP = None
    # where do we keep unstable arrow files
    TMPDIR = None
    # filter variant matches to a single chromosome?
    CHROM = None
    # scorefile path
    SCOREFILE = None
    # clean up temporary arrow files created by context manager?
    CLEANUP = False
    # scorefile output options
    SPLIT = False
    COMBINED = False
    # matching parameter arguments
    MATCH_KWARGS = [
        "keep_first_match",
        "remove_ambiguous",
        "skip_flip",
        "filter_IDs",
        "remove_multiallelic",
    ]
    MATCH_PARAMS = None
    MIN_OVERLAP = 0.75



import logging

from .lib import VariantFrame
from .lib import ScoringFileFrame, match_variants
from .lib import MatchResult, MatchResults
from .lib import PlinkScoreFiles


log_fmt = "%(name)s: %(asctime)s %(levelname)-8s %(message)s"
logging.basicConfig(format=log_fmt, datefmt="%Y-%m-%d %H:%M:%S")
logger = logging.getLogger(__name__)

__all__ = [
    "VariantFrame",
    "ScoringFileFrame",
    "MatchResult",
    "MatchResults",
    "PlinkScoreFiles",
    "match_variants",
]

__version__ = "0.2.0"



"""This module contains internal functions for writing match results to scoring files and logs

It expects Config class attributes to be set up before being called
"""

import gzip
import itertools

from pgscatalog.core import MatchRateError

from ..lib.plinkscorefiles import PlinkScoreFiles

from ._config import Config


def write_matches(matchresults, score_df):
    """Write matchresults out to scoring files and logs"""
    try:
        match (Config.SPLIT, Config.COMBINED):
            case (True, True):
                # requires extra work: first write split
                outfs = matchresults.write_scorefiles(
                    directory=Config.OUTDIR,
                    split=True,
                    score_df=score_df,
                    min_overlap=Config.MIN_OVERLAP,
                    **Config.MATCH_PARAMS,
                )
                # now re-combine without recomputing matches
                PlinkScoreFiles(*list(itertools.chain(*outfs))).merge(Config.OUTDIR)
            case (True, False) | (False, True):
                # split parameter can handle this case OK
                _ = matchresults.write_scorefiles(
                    directory=Config.OUTDIR,
                    split=Config.SPLIT,
                    score_df=score_df,
                    min_overlap=Config.MIN_OVERLAP,
                    **Config.MATCH_PARAMS,
                )
            case _:
                raise ValueError
    except MatchRateError as e:
        # write logs in case of errors
        write_log(matchresults=matchresults, score_df=score_df)
        raise e
    else:
        write_log(matchresults=matchresults, score_df=score_df)

    # returns labelled and filtered data for checking after merging
    return matchresults.df


def write_log(matchresults, score_df):
    logfname = Config.OUTDIR / f"{Config.DATASET}_log.csv.gz"

    # summary log is smol
    matchresults.summary_log.write_csv(Config.OUTDIR / f"{Config.DATASET}_summary.csv")

    # this one can get big. gzip is slow, but everywhere
    with gzip.open(logfname, "wb") as f:
        matchresults.full_variant_log(score_df=score_df).collect().write_csv(f)



""" This module contains functions for streaming text files to arrow IPC files

It reimplements CSV reading using the polars API directly to be fast.

pgscatalog.core's implementation with csv.DictReader is slow with hundreds of millions of lines.
"""
import pathlib
import tempfile

import polars as pl
from functools import singledispatch

import polars.exceptions
from pgscatalog.core import NormalisedScoringFile, TargetVariants, TargetType


@singledispatch
def loose(path, tmpdir=None):
    """Loose an arrow :) Stream compressed text files into temporary arrow files

    This approach relies heavily on polars to read and write tabular data efficiently
    """
    # can currently only handle ScoringFile and TargetVariants
    raise NotImplementedError


@loose.register
def _(path: NormalisedScoringFile, tmpdir=None):
    """Write NormalisedScoringFiles to a list of arrow IPC files"""
    if tmpdir is None:
        tmpdir = tempfile.mkdtemp()
    # alleles are first read as strings and converted to categoricals later
    dtypes = {
        "chr_name": pl.Categorical,
        "chr_position": pl.UInt64,
        "effect_allele": pl.String,
        "other_allele": pl.String,
        "effect_weight": pl.String,
        "effect_type": pl.Categorical,
        "is_duplicated": pl.Boolean,
        "accession": pl.Categorical,
        "row_nr": pl.UInt64,
    }

    reader = pl.read_csv_batched(path.path, dtypes=dtypes, separator="\t")
    return batch_read(reader, tmpdir=tmpdir, cols_keep=pl.all())


@loose.register
def _(path: TargetVariants, tmpdir=None):
    """Writes TargetVariants to a list of arrow IPC files"""
    if tmpdir is None:
        tmpdir = tempfile.mkdtemp()

    match path.ftype:
        case TargetType.BIM:
            # no header
            dtypes = {
                "column_1": pl.Categorical,
                "column_2": pl.String,
                "column_3": pl.UInt8,
                "column_4": pl.UInt64,
                "column_5": pl.String,
                "column_6": pl.String,
            }
            new_colnames = ["#CHROM", "ID", "CM", "POS", "REF", "ALT"]
            header = False
            comment = None
        case TargetType.PVAR:
            # order of dict is important
            dtypes = {
                "#CHROM": pl.Categorical,
                "POS": pl.UInt64,
                "ID": pl.String,
                "REF": pl.String,
                "ALT": pl.String,
            }
            header = True
            comment = "##"
            new_colnames = None  # don't rename, use the header
        case _:
            raise NotImplementedError

    reader = pl.read_csv_batched(
        path.path,
        has_header=header,
        dtypes=dtypes,
        separator="\t",
        new_columns=new_colnames,
        comment_prefix=comment,
        batch_size=1000000,
    )
    cols_keep = ["#CHROM", "POS", "ID", "REF", "ALT"]
    return batch_read(reader, tmpdir=tmpdir, cols_keep=cols_keep)


def batch_read(reader, tmpdir, cols_keep) -> list[pathlib.Path]:
    """Read a CSV in batches and write them to temporary files"""
    arrowpaths = []
    # batch_size should be >= thread pool size, so tasks will be distributed amongst workers
    batch_size = 100
    batches = reader.next_batches(batch_size)
    while batches:
        arrowpath = tempfile.NamedTemporaryFile(dir=tmpdir, delete=False)
        df = pl.concat(batches).select(cols_keep)
        df.write_ipc(arrowpath)
        batches = reader.next_batches(batch_size)
        arrowpaths.append(pathlib.Path(arrowpath.name))

    return arrowpaths



import logging

import polars as pl

logger = logging.getLogger(__name__)


def make_logs(scorefile: pl.LazyFrame, match_candidates: pl.LazyFrame, dataset: str):
    # summary log -> aggregated from best matches (one per scoring file line)
    # big log -> unaggregated, written to compressed gzip, possibly multiple matches per scoring file line
    big_log = _join_match_candidates(
        scorefile=scorefile, matches=match_candidates, dataset=dataset
    )

    return _prettify_log(big_log)


def make_summary_log(
    match_candidates: pl.LazyFrame,
    scorefile: pl.LazyFrame,
    filter_summary: pl.LazyFrame,
    dataset: str,
) -> pl.LazyFrame:
    """Make an aggregated table that contains the best match candidate for each row in the original scoring file"""
    logger.debug("Aggregating best matches into a summary table")
    cols = [
        "dataset",
        "accession",
        "ambiguous",
        "is_multiallelic",
        "match_flipped",
        "duplicate_best_match",
        "duplicate_ID",
        "match_IDs",
        "match_status",
    ]

    best_matches: pl.LazyFrame = match_candidates.filter(pl.col("best_match"))

    return (
        scorefile.join(best_matches, on=["row_nr", "accession"], how="full")
        .with_columns(
            pl.col("match_status").fill_null(value="unmatched"), dataset=pl.lit(dataset)
        )  # fill in unmatched variants
        .group_by(cols)
        .len()
        .rename({"len": "count"})
        .join(filter_summary, how="left", on="accession")
        .pipe(_prettify_summary)
    )


def check_log_count(scorefile: pl.LazyFrame, summary_log: pl.LazyFrame):
    """Check aggregated counts vs original from scoring file"""
    summary_count: pl.LazyFrame = summary_log.group_by("accession").agg(pl.sum("count"))

    log_count: pl.DataFrame = (
        scorefile.group_by("accession")
        .len()
        .rename({"len": "count"})
        .join(summary_count, on="accession")
        .collect()
    )

    for row in log_count.iter_rows():
        if (log_count := row[1]) != (score_count := row[2]):
            logger.critical("Variant log doesn't match input scoring file counts")
            raise ValueError(
                f"{row[0]} match failure {log_count=} doesn't match {score_count=}"
            )

    return True


def _prettify_summary(df: pl.LazyFrame) -> pl.LazyFrame:
    keep_cols = [
        "dataset",
        "accession",
        "score_pass",
        "match_status",
        "ambiguous",
        "is_multiallelic",
        "duplicate_best_match",
        "duplicate_ID",
        "match_flipped",
        "match_IDs",
        "count",
        "percent",
    ]
    return df.with_columns(
        (pl.col("count") / pl.sum("count") * 100)
        .over(["dataset", "accession"])
        .alias("percent")
    ).select(keep_cols)


def _prettify_log(df: pl.LazyFrame) -> pl.LazyFrame:
    keep_cols = [
        "row_nr",
        "accession",
        "chr_name",
        "chr_position",
        "effect_allele",
        "other_allele",
        "effect_weight",
        "effect_type",
        "ID",
        "REF",
        "ALT",
        "matched_effect_allele",
        "match_type",
        "is_multiallelic",
        "ambiguous",
        "match_flipped",
        "best_match",
        "exclude",
        "duplicate_best_match",
        "duplicate_ID",
        "match_IDs",
        "match_status",
        "dataset",
    ]
    pretty_df = df.select(keep_cols).sort(
        ["accession", "row_nr", "chr_name", "chr_position", "match_status"]
    )
    return pretty_df


def _join_match_candidates(
    scorefile: pl.LazyFrame, matches: pl.LazyFrame, dataset: str
) -> pl.LazyFrame:
    """Join match candidates against the original scoring file"""
    logger.debug("Joining all match candidates against input scoring file")
    # make a raw log with all match candidates included
    raw_log = (
        scorefile.join(matches, on=["row_nr", "accession"], how="full")
        .with_columns(pl.lit(dataset).alias("dataset").cast(pl.Categorical))
        .select(pl.exclude("^.*_right$"))
    ).with_columns(pl.col("match_status").fill_null("unmatched"))

    return raw_log



"""This module contains functions to filter match candidates and report overall match rates"""

import logging

import polars as pl

logger = logging.getLogger(__name__)


def filter_scores(
    scorefile: pl.LazyFrame, matches: pl.LazyFrame, min_overlap: float, dataset: str
) -> tuple[pl.LazyFrame, pl.DataFrame]:
    """Check overlap between filtered matches and scorefile

    Return a df with bad scores removed and a summary table"""
    filtered_matches: pl.LazyFrame = _filter_matches(matches)
    match_log: pl.LazyFrame = _join_filtered_matches(
        filtered_matches, scorefile, dataset
    ).with_columns(pl.col("best_match").fill_null(False))

    fail_rates: pl.DataFrame = _calculate_match_rate(
        match_log
    ).collect()  # collect for iteration

    scores: list[pl.DataFrame] = []
    for accession, rate in zip(
        fail_rates["accession"].to_list(), fail_rates["fail_rate"].to_list()
    ):
        if rate <= (1 - min_overlap):
            df: pl.DataFrame = pl.DataFrame(
                {
                    "accession": [accession],
                    "score_pass": [True],
                    "match_rate": [1 - rate],
                }
            )
            logger.debug(
                f"Score {accession} passes minimum matching threshold ({1 - rate:.2%}  variants match)"
            )
            scores.append(df.with_columns(pl.col("accession").cast(pl.Categorical)))
        else:
            df: pl.DataFrame = pl.DataFrame(
                {
                    "accession": [accession],
                    "score_pass": [False],
                    "match_rate": [1 - rate],
                }
            )
            logger.error(
                f"Score {accession} fails minimum matching threshold ({1 - rate:.2%} variants match)"
            )
            scores.append(df.with_columns(pl.col("accession").cast(pl.Categorical)))

    score_summary: pl.LazyFrame = pl.concat(scores).lazy()
    filtered_scores: pl.LazyFrame = filtered_matches.join(
        score_summary, on="accession", how="left"
    ).filter(pl.col("score_pass"))

    return filtered_scores, score_summary.collect()


def _calculate_match_rate(df: pl.LazyFrame) -> pl.LazyFrame:
    logger.debug("Calculating overlap between target genome and scoring file")
    return (
        df.group_by("accession")
        .agg([pl.all().len(), (pl.col("match_type").is_null()).sum().alias("no_match")])
        .with_columns((pl.col("no_match") / pl.col("count")).alias("fail_rate"))
    )


def _filter_matches(df: pl.LazyFrame) -> pl.LazyFrame:
    logger.debug("Filtering to best_match variants (with exclude flag = False)")
    return df.filter((pl.col("best_match")) & (~pl.col("exclude")))


def _join_filtered_matches(
    matches: pl.LazyFrame, scorefile: pl.LazyFrame, dataset: str
) -> pl.LazyFrame:
    return (
        scorefile.join(matches, on=["row_nr", "accession"], how="left")
        .with_columns(pl.lit(dataset).cast(pl.Categorical).alias("dataset"))
        .select(pl.exclude("^.*_right$"))
    )


def _match_keys() -> list[str]:
    return [
        "chr_name",
        "chr_position",
        "effect_allele",
        "other_allele",
        "accession",
        "effect_type",
        "effect_weight",
    ]



import argparse
import atexit
import logging
import pathlib
import shutil

from ._config import Config
from ._write import write_matches

from ..lib import ScoringFileFrame, MatchResult, MatchResults

import polars as pl

logger = logging.getLogger(__name__)


def _exit(cleanup):
    if cleanup:
        if Config.TMPDIR and Config.TMPDIR.exists():
            shutil.rmtree(Config.TMPDIR)
        if Config.MATCHTMP and Config.MATCHTMP.exists():
            shutil.rmtree(Config.MATCHTMP)


def run_merge():
    args = parse_args()

    if args.verbose:
        logging.getLogger("pgscatalog.core").setLevel(logging.DEBUG)
        logging.getLogger("pgscatalog.match").setLevel(logging.DEBUG)
        logger.setLevel(logging.DEBUG)
        logger.debug("Verbose logging enabled")

    Config.DATASET = args.dataset
    Config.CLEANUP = args.cleanup
    Config.OUTDIR = pathlib.Path(args.outdir)
    Config.SCOREFILE = pathlib.Path(args.scorefile)

    if not Config.OUTDIR.exists():
        raise FileNotFoundError(f"{Config.OUTDIR} does not exist")

    atexit.register(_exit, cleanup=Config.CLEANUP)

    if Config.CLEANUP:
        logger.info("--cleanup set (default), temporary files will be deleted")
    else:
        logger.info("--no-cleanup set, temporary files won't be deleted")

    Config.TMPDIR = Config.OUTDIR / "tmp"
    Config.TMPDIR.mkdir(exist_ok=False)

    # parameters that control how the best match candidate is chosen
    # missing parameters will be set to defaults specified in matchlib
    Config.MATCH_PARAMS = {
        k: v for k in Config.MATCH_KWARGS if (v := getattr(args, k)) is not None
    }
    Config.MIN_OVERLAP = args.min_overlap
    Config.SPLIT = args.split
    Config.COMBINED = args.combined

    with ScoringFileFrame(
        path=Config.SCOREFILE,
        chrom=None,  # when merging, scoring files can't be filtered
        cleanup=Config.CLEANUP,
        tmpdir=Config.TMPDIR,
    ) as score_df, pl.StringCache():
        matchresults = MatchResults(
            *(MatchResult.from_ipc(x, dataset=Config.DATASET) for x in args.matches)
        )
        matchdf = write_matches(matchresults=matchresults, score_df=score_df)
        _check_duplicate_vars(matchdf)
    logger.info("finished merging :)")


def _check_duplicate_vars(matches):
    max_occurrence = (
        matches.filter(pl.col("match_status") == "matched")
        .group_by(["accession", "ID"])
        .len()
        .select("len")
        .max()
        .collect()
        .item(0, 0)
    )

    match n := max_occurrence:
        case None:
            logger.critical("No variant matches found")
            logger.critical(
                "Did you set the correct genome build? Did you impute your genomes?"
            )
            raise ValueError
        case _ if n > 1:
            logger.critical("Duplicate IDs in final matches")
            logger.critical(
                "Please double check your genomes for duplicates and try again"
            )
            raise ValueError
        case _:
            logger.info("Scoring files are valid (no duplicate variants found)")


def parse_args(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d",
        "--dataset",
        dest="dataset",
        required=True,
        help="<Required> Label for target genomic dataset",
    )
    parser.add_argument(
        "-s",
        "--scorefile",
        dest="scorefile",
        required=True,
        help="<Required> Path to scorefile",
    )
    parser.add_argument(
        "-m",
        "--matches",
        dest="matches",
        required=True,
        nargs="+",
        help="<Required> List of match files",
    )
    parser.add_argument(
        "--min_overlap",
        dest="min_overlap",
        required=True,
        type=float,
        default=0.75,
        help="<Required> Minimum proportion of variants to match before error",
    )
    parser.add_argument(
        "-IDs",
        "--filter_IDs",
        dest="filter_IDs",
        help="<Optional> Path to file containing list of variant IDs that can be included in the final scorefile."
        "[useful for limiting scoring files to variants present in multiple datasets]",
    )
    parser.add_argument(
        "--outdir", dest="outdir", required=True, help="<Required> Output directory"
    )
    parser.add_argument(
        "--split",
        dest="split",
        default=False,
        action="store_true",
        help="<Optional> Write scorefiles split per chromosome?",
    )
    parser.add_argument(
        "--combined",
        dest="combined",
        default=False,
        action="store_true",
        help="<Optional> Write scorefiles in combined format?",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        help="<Optional> Extra logging information",
    )
    parser.add_argument(
        "--cleanup", dest="cleanup", default=True, action=argparse.BooleanOptionalAction
    )
    # variant matching arguments -------------------------------------------------------
    parser.add_argument(
        "--keep_ambiguous",
        dest="remove_ambiguous",
        action="store_false",
        help="""<Optional> Flag to force the program to keep variants with
                        ambiguous alleles, (e.g. A/T and G/C SNPs), which are normally
                        excluded (default: false). In this case the program proceeds
                        assuming that the genotype data is on the same strand as the
                        GWAS whose summary statistics were used to construct the score.
                                """,
    )
    parser.add_argument(
        "--keep_multiallelic",
        dest="remove_multiallelic",
        action="store_false",
        help="<Optional> Flag to allow matching to multiallelic variants (default: false).",
    )
    parser.add_argument(
        "--ignore_strand_flips",
        dest="skip_flip",
        action="store_true",
        help="""<Optional> Flag to not consider matched variants that may be reported 
                        on the opposite strand.  Default behaviour is to flip/complement unmatched variants and check if
                        they match.""",
    )
    parser.add_argument(
        "--keep_first_match",
        dest="keep_first_match",
        action="store_true",
        help="""<Optional> If multiple match candidates for a variant exist that can't be prioritised,
                         keep the first match candidate (default: drop all candidates)""",
    )

    return _check_args(parser.parse_args(args))


def _check_args(args):
    if args.combined is False and args.split is False:
        logger.warning("No output format specified, writing to combined scoring file")
        args.combined = True

    return args


if __name__ == "__main__":
    run_merge()
